PLAYER_BLACK = 0
PLAYER_WHITE = 1

GAME_WIDTH = 600
GAME_HEIGHT = 600
STATUS_HEIGHT = 50

MAX_POWER = 2500
GRID_SIZE = 600 // 20
DAMPING = 0.6  # Velocity damping factor (lower value = more damping)
VELOCITY_THRESHOLD = 5.0  # Threshold below which stones are considered "at rest"